# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Kevin De Bruycker and Tim Krappitz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import absolute_import
from collections import Counter
import itertools
import pymacroms
import matplotlib.pyplot as plt
import pandas as pd
from operator import itemgetter, attrgetter
import re
import sys
import progressbar

class Polymer:

    def __init__(self,
                 endgroupPairs,
                 monomers,
                 adductIon = "H+",
                 charge = 1,
                 minRelAbundance = 0.01,
                 minRepeatingUnits = None,
                 maxRepeatingUnits = None,
                 mzRange = None,
                 customEndgroupsDatabase = None,
                 customMonomersDatabase = None,
                 firstMonomerMajor: bool = False,
                 comonomersToFirstMaxRatio: float = 1,
                 lastMonomerMinor: bool = False):
        """

        :type endgroupPairs: list
        :type monomers: list
        :type adductIon: str
        :type charge: int
        :type minRelAbundance: float
        :type minRepeatingUnits: int or list
        :type maxRepeatingUnits: int or list
        :type mzRange: list
        :type customEndgroupsDatabase: dict
        :type customMonomersDatabase: dict
        """

        # Internal formulas of endgroups and monomers should be (lists of) Counters for processing
        # Their self.attr counterpart is then the counter printed as str for easy external readout.
        #todo Maybe allow to enter chemical formulas directly, but then the code should first check whether the string is found in the database and if not transform it using the getCounterFromFormula function. Not urgent

        # input processing
        if customEndgroupsDatabase is not None:
            self.customEndgroupsDatabase = customEndgroupsDatabase
        else:
            self.customEndgroupsDatabase = dict()
        if customMonomersDatabase is not None:
            self.customMonomersDatabase = customMonomersDatabase
        else:
            self.customMonomersDatabase = dict()

        print("Simulating polymer...")
        if type(endgroupPairs) != list:
            endgroupPairs = [endgroupPairs]
        self.endgroupPairs = []
        for index, endgroupPair in enumerate(endgroupPairs):
            if type(endgroupPair) == str: # lookup the entry in the endgroups list in the database
                self.endgroupPairs.append(endgroupPair)
                if endgroupPair in self.customEndgroupsDatabase:
                    endgroupPairs[index] = self.customEndgroupsDatabase[endgroupPair][0] + self.customEndgroupsDatabase[endgroupPair][1]
                elif endgroupPair in pymacroms.database.endgroups:
                    endgroupPairs[index] = pymacroms.database.endgroups[endgroupPair][0] + pymacroms.database.endgroups[endgroupPair][1]
                else:
                    sys.exit("endgroupPair " + endgroupPair + " not found in either custom or built-in database!")
            elif type(endgroupPair) == list: # entry in the endgroups list is a list --> both endgroups are separate so should be combined first
                endgroupPairs[index] = endgroupPair[0] + endgroupPair[1]
                self.endgroupPairs.append(pymacroms.getFormulaStrFromCounter(endgroupPairs[index]))
            else: # entry is a counter, 1 counter for the 2 endgroups of the polymer
                self.endgroupPairs.append(pymacroms.getFormulaStrFromCounter(endgroupPair))
        if type(monomers) != list:
            monomers = [monomers]
        self.monomers = []
        for index, monomer in enumerate(monomers):
            if type(monomer) == str:
                self.monomers.append(monomer)
                if monomer in self.customMonomersDatabase:
                    monomers[index] = self.customMonomersDatabase[monomer]
                elif monomer in pymacroms.database.monomers:
                    monomers[index] = pymacroms.database.monomers[monomer]
                else:
                    sys.exit("Monomer " + monomer + " not found in either custom or built-in database!")
            else:
                self.monomers.append(pymacroms.getFormulaStrFromCounter(monomer))

        self.adductIon = adductIon
        self.charge = charge
        try:
            adductIon = pymacroms.database.ionising_species[adductIon]  # [formula: Counter, charge: int, mass_correction: float]
        except:
            sys.exit("adductIon not found in database!")
        self.amountAdductIons = abs(int(charge / adductIon[1]))
        # macromolecules will be generated with a defined minimum and maximum (total) number of repeating units
        # These boundaries can be given as arguments, but can also be estimated from an mzRange and
        # the molecular weights of the smallest (maxRU) and biggest (minRU) monomer:
        if mzRange is None and maxRepeatingUnits is None:
            # it will be impossible to estimate the range of repeating units
            sys.exit("I'm struggling to build a matrix of possible combinations of monomers based on \nthe provided combination of mzRange, minRepeatingUnits and maxRepeatingUnits.\nPlease check the parameters.")
        elif maxRepeatingUnits is None:
            if minRepeatingUnits is None:
                minRepeatingUnits = max(int((mzRange[0] * abs(charge) - max(list(pymacroms.getMonoIsotopicMass(endgroupPair) for endgroupPair in endgroupPairs)) - abs(int(charge / adductIon[1])) * pymacroms.getMonoIsotopicMass(adductIon[0])) / max(list(pymacroms.getMonoIsotopicMass(monomer) for monomer in monomers))), 0)
                print("--> minimum sum of repeating units: %i (estimated based on the provided mzRange and the masses of the components)" % minRepeatingUnits)
            maxRepeatingUnits = int((mzRange[1] * abs(charge) - min(list(pymacroms.getMonoIsotopicMass(endgroupPair) for endgroupPair in endgroupPairs)) - abs(int(charge / adductIon[1])) * pymacroms.getMonoIsotopicMass(adductIon[0])) / min(list(pymacroms.getMonoIsotopicMass(monomer) for monomer in monomers))) + 1
            print("--> maximum sum of repeating units: %i (estimated based on the provided mzRange and the masses of the components)" % maxRepeatingUnits)
        elif minRepeatingUnits is None:
            if mzRange is not None:
                minRepeatingUnits = max(int((mzRange[0] * abs(charge) - max(list(pymacroms.getMonoIsotopicMass(endgroupPair) for endgroupPair in endgroupPairs)) - abs(int(charge / adductIon[1])) * pymacroms.getMonoIsotopicMass(adductIon[0])) / max(list(pymacroms.getMonoIsotopicMass(monomer) for monomer in monomers))), 0)
                print("--> minimum sum of repeating units: %i (estimated based on the provided mzRange and the masses of the components)" % minRepeatingUnits)
            else:
                minRepeatingUnits = 0
                print("--> minimum sum of repeating units set to 0 since it cannot be estimated")
        # Now build all the combinations of repeating units within the limits of the given (or estimated) parameters
        if type(minRepeatingUnits) == int:
            # in this case, only a global minimum is given, so all combinations have to be made from 0
            # the ones with a sum smaller than minRepeatingUnits will be filtered later
            if type(maxRepeatingUnits) == int:
                # most simple case: every monomer from 0 to the same maxRepeatingUnits
                # combinations with a too high sum will be filtered later
                RUCombinations = list(itertools.product(range(maxRepeatingUnits + 1), repeat=len(monomers)))
            elif type(maxRepeatingUnits) == list:
                # first check that the length of the list is valid
                if len(maxRepeatingUnits) != len(monomers):
                    sys.exit("The length of maxRepeatingUnits does not match the amount of monomers!\nEither provide a max for each monomer or 1 global max for the sum of the repeating units")
                # similar as before but instead of the same range repeated for "amount of monomer times", here different ranges are used
                RUCombinations = list(itertools.product(*(range(max + 1) for max in maxRepeatingUnits)))
                maxRepeatingUnits = sum(maxRepeatingUnits)
            else:
                sys.exit("Something went wrong when interpreting maxRepeatingUnits\nCheck your input")
        elif type(minRepeatingUnits) == list:
            # first check that the length of the list is valid
            if len(minRepeatingUnits) != len(monomers):
                sys.exit("The length of minRepeatingUnits does not match the amount of monomers!\nEither provide a min for each monomer or 1 global min for the sum of the repeating units")
            if type(maxRepeatingUnits) == int:
                # the same formula as earlier can be used, but now the list comprehension will be over the minimum with a constant maximum
                # again, the combinations with a too high sum will be filtered later
                RUCombinations = list(itertools.product(*(range(min, maxRepeatingUnits + 1) for min in minRepeatingUnits)))
            elif type(maxRepeatingUnits) == list:
                # first check that the length of the list is valid
                if len(maxRepeatingUnits) != len(monomers):
                    sys.exit("The length of maxRepeatingUnits does not match the amount of monomers!\nEither provide a max for each monomer or 1 global max for the sum of the repeating units")
                # now both the minima and maxima are fixed, both min and max lists are combined using zip
                RUCombinations = list(itertools.product(*(range(min, max + 1) for min, max in list(zip(minRepeatingUnits, maxRepeatingUnits)))))
                maxRepeatingUnits = sum(maxRepeatingUnits)
            else:
                sys.exit("Something went wrong when interpreting maxRepeatingUnits\nCheck your input")
            minRepeatingUnits = sum(minRepeatingUnits)
        else:
            sys.exit("Something went wrong when interpreting minRepeatingUnits\nCheck your input")
        if minRepeatingUnits > maxRepeatingUnits:
            sys.exit("With the current parameters, minRepeatingUnits is higher than maxRepeatingUnits, this can't be correct...")
        # Finally, combine the combinations of repeating units with the different indices of the different end groups
        compositions = list(itertools.product(range(len(endgroupPairs)), RUCombinations))
        print("--> Possible combinations of repeating units/end groups based on given parameters: " + str(len(compositions)))
        print("--> Filtering and calculating the isotopic patterns...")

        formulaDatabase = [[],[]]
        macromolecules = []
        isobaricSpecies = False
        for EGPairIndex, RUCombination in compositions:
            if sum(RUCombination) < minRepeatingUnits or sum(RUCombination) > maxRepeatingUnits: # kick out the monomer combinations that result in a total amount of repeating units outside the defined range
                continue
            if firstMonomerMajor and RUCombination[0] < (sum(RUCombination) - RUCombination[0])/comonomersToFirstMaxRatio:
                continue
            if lastMonomerMinor and len(RUCombination) > 2 and RUCombination[len(RUCombination)-1] > sum(RUCombination) - RUCombination[0] - RUCombination[len(RUCombination)-1]: #
                # only makes sense when there are 3 or more different monomers
                # will still include the combination that has the same amount of last monomer as the sum of others
                # RUCombination[len(RUCombination)-1] > min(RUCombination) is bad plan
                continue
            formulaCounter = Counter()
            for index, numberRU in enumerate(RUCombination):
                formulaCounter += Counter({element:amount*numberRU for element,amount in monomers[index].items()})
            formulaCounter = endgroupPairs[EGPairIndex] + formulaCounter
            tempMolecule = pymacroms.Molecule(formulaCounter, self.adductIon, charge, minRelAbundance)
            # kick out the molecule if it's outside the defined mass range
            if mzRange is not None:
                if not tempMolecule.inMassRange(mzRange, False):
                    continue
            if formulaCounter in formulaDatabase[0]:
                formulaDatabase[1][formulaDatabase[0].index(formulaCounter)].append([RUCombination, EGPairIndex])
                isobaricSpecies = True
            else:
                formulaDatabase[0].append(formulaCounter)
                formulaDatabase[1].append([[RUCombination, EGPairIndex]])
                macromolecules.append({"repUnitsTotal": sum(RUCombination), "repUnitsCombination": RUCombination, "endgroupPairIndex": EGPairIndex, "moleculeData": tempMolecule})
        if isobaricSpecies:
            self.isobaricSpecies = []
            for index, [counter, data] in enumerate(zip(formulaDatabase[0], formulaDatabase[1])):
                if len(data) > 1:
                    self.isobaricSpecies.append({"repUnitsCombinations": list(zip(*data))[0], "endgroupPairIndices": list(zip(*data))[1], "moleculeData": macromolecules[index]["moleculeData"]})
        else:
            self.isobaricSpecies = None
        print("--> Number of (non-isobaric) macromolecules retained based on restrictions: " + str(len(macromolecules)))
        self.macromolecules = sorted(macromolecules, key=itemgetter("repUnitsTotal"))
        self.isotopicDist = self.getIsotopicDist()
        self.mzRange = [self.isotopicDist[0][0], self.isotopicDist[len(self.isotopicDist) - 1][0]]
        print("--> Done\n")

    def aggregateIsotopes(self, resolution: float):
        print("--> Combining isotopologues within resolution limits for all %i macromolecules" % len(self.macromolecules))
        sys.stdout.flush()
        if hasattr(self.macromolecules[0]["moleculeData"], "resolution"):
            if self.macromolecules[0]["moleculeData"].resolution == resolution:
                print("--> Isotopologues have already been aggregated with this resolution")
                return
        for macromolecule in progressbar.progressbar(self.macromolecules, 0, len(self.macromolecules)):
            macromolecule["moleculeData"].resolution = resolution
            macromolecule["moleculeData"].isotopicDist_resolution = pymacroms.combineIsotopes(macromolecule["moleculeData"].isotopicDist, resolution)
            macromolecule["moleculeData"].massMostAbundant = sorted(macromolecule["moleculeData"].isotopicDist_resolution, key=itemgetter(1), reverse=True)[0][0]

    def getIsotopicDist(self, resolution: float = None):
        isotopicDist = []
        if resolution is not None:
            self.aggregateIsotopes(resolution)
        for macromolecule in self.macromolecules:
            if resolution is None:
                for mass, normAbund in macromolecule["moleculeData"].isotopicDist:
                    isotopicDist.append([mass, normAbund])
            else:
                for mass, normAbund in macromolecule["moleculeData"].isotopicDist_resolution:
                    isotopicDist.append([mass, normAbund])
        return sorted(isotopicDist)

    def getMoleculeList(self, experimentalMass: float = None, ppmDev: float = 5, resolution: float = None):
        isotopesNearMass = None
        moleculeList = []
        for index, macromolecule in enumerate(self.macromolecules):
            if experimentalMass is not None:
                isotopesNearMass = macromolecule["moleculeData"].getIsotopesNearMass(experimentalMass, ppmDev, resolution)
                if isotopesNearMass is None:
                    continue
            moleculeList.append([index, list(macromolecule["repUnitsCombination"]), isotopesNearMass])
        # return sorted(moleculeList, key=itemgetter(2))
        if len(moleculeList) > 0:
            return sorted(moleculeList, key=itemgetter(1))
        else:
            return None

    # def printMoleculeList(self, experimentalMass: float = None, ppmDev: float = 5):
    #     print("Monomer combination\tFormula (ion)\tMost abundant mass")
    #     # print("\t" + str(self.monomers))
    #     for macromoleculeIndex, repUnitsCombination, formula_ion, massMostAbundant in self.getMoleculeList(experimentalMass, ppmDev):
    #         print(str(repUnitsCombination) + "\t" + formula_ion + "\t" + str(massMostAbundant))

    def printMoleculeList(self, experimentalMass: float = None, ppmDev: float = 5, resolution: float = None):
        moleculeList = self.getMoleculeList(experimentalMass, ppmDev, resolution)
        if moleculeList is not None:
            print("Experimental mass: " + str(experimentalMass))
            print("Monomer combination\tEnd-groups\tFormula (ion)\tMost abundant mass\tMatched isotope(s) (m/z, dev. (ppm))")
            # print(str(self.monomers) + "\t\t\t(m/z, norm. abundance, dev (ppm))")
            for macromoleculeIndex, repUnitsCombination, isotopesNearMass in moleculeList:
                if isotopesNearMass is not None:
                    print(str(repUnitsCombination) + "\t" + self.endgroupPairs[self.macromolecules[macromoleculeIndex]["endgroupPairIndex"]] + "\t" + self.macromolecules[macromoleculeIndex]["moleculeData"].formula_ion + "\t" + str(self.macromolecules[macromoleculeIndex]["moleculeData"].massMostAbundant) + "\t" + str(list((mz, round(dev, 2)) for mz, normAbund, dev in isotopesNearMass)))
                else:
                    print(str(repUnitsCombination) + "\t" + self.endgroupPairs[
                        self.macromolecules[macromoleculeIndex]["endgroupPairIndex"]] + "\t" +
                          self.macromolecules[macromoleculeIndex]["moleculeData"].formula_ion + "\t" + str(
                        self.macromolecules[macromoleculeIndex]["moleculeData"].massMostAbundant) + "\tN/A")
        else:
            print("No matching molecules found...")

    def saveMoleculeList(self, filename: str, experimentalMass: float = None, ppmDev: float = 5, resolution: float = None):
        moleculeList = self.getMoleculeList(experimentalMass, ppmDev, resolution)
        if moleculeList is not None:
            output = []
            for macromoleculeIndex, repUnitsCombination, isotopesNearMass in moleculeList:
                if isotopesNearMass is not None:
                    output.append([str(repUnitsCombination), self.endgroupPairs[self.macromolecules[macromoleculeIndex]["endgroupPairIndex"]], self.macromolecules[macromoleculeIndex]["moleculeData"].formula_ion, str(self.macromolecules[macromoleculeIndex]["moleculeData"].massMostAbundant), str(list((mz, round(dev, 2)) for mz, normAbund, dev in isotopesNearMass))])
                else:
                    output.append([str(repUnitsCombination), self.endgroupPairs[self.macromolecules[macromoleculeIndex]["endgroupPairIndex"]], self.macromolecules[macromoleculeIndex]["moleculeData"].formula_ion, str(self.macromolecules[macromoleculeIndex]["moleculeData"].massMostAbundant), "N/A"])
            export = pd.DataFrame(output,
                                  columns=["Monomer combination", "End-groups", "Formula (ion)", "Most abundant mass",
                                           "Matched isotope(s) (m/z, dev. (ppm))"])
        else:
            export = pd.DataFrame(columns=["No matching molecules found..."])
        export.to_csv(filename + ".csv", index=False)

    def printFullInfo(self):
        print("Total number of repeating units\tMonomer combination\tEnd-groups\tFormula (ion)\tm/z\tRel. Abundance")
        print("\t" + str(self.monomers))
        for macromolecule in self.macromolecules:
            i = 1
            for mass, relAbund in pymacroms.toRelativeAbundance(macromolecule["moleculeData"].isotopicDist):
                if i == 1:
                    print(str(macromolecule["repUnitsTotal"]) + "\t" + str(list(macromolecule["repUnitsCombination"])) + "\t" + self.endgroupPairs[macromolecule["endgroupPairIndex"]] + "\t" + macromolecule["moleculeData"].formula_ion + "\t" + str(mass) + "\t" + str(relAbund))
                else:
                    print("\t\t\t\t" + str(mass) + "\t" + str(relAbund))
                i += 1

    def printIsotopicDist(self, resolution: float = None):
        print("Mass\tRel. Abundance")
        print("\t" + str(self.monomers))
        if resolution is not None:
            if not hasattr(self, "isotopicDist_resolution"):
                self.isotopicDist_resolution = self.getIsotopicDist(resolution)
            isotopicDist_rel = pymacroms.toRelativeAbundance(self.isotopicDist_resolution)
        else:
            isotopicDist_rel = pymacroms.toRelativeAbundance(self.isotopicDist)
        for mass, relAbund in isotopicDist_rel:
            print(str(mass) + "\t" + str(relAbund))

    def plotIsotopicDist(self, color = "black", mzRange: list = None):
        mz_axis = []
        relAbund_axis = []
        for mz, relAbund in pymacroms.toRelativeAbundance(self.isotopicDist):
            mz_axis += list([mz - 1e-10, mz, mz + 1e-10])
            relAbund_axis += list([-0.1, relAbund, -0.1])
        plt.plot(mz_axis, relAbund_axis, color=color)
        plt.title(r"Isotopic distribution ({}{} adducts) of a (co)polymer of".format(self.amountAdductIons, re.sub("([0-9]+)", "$_{\\1}$", re.sub("([+-]+)", "$^{\\1}$", self.adductIon))) + "\n" +
                  "{} with end-group(s): {}".format(re.sub("([0-9]+)", "$_{\\1}$", str(self.monomers)), re.sub("([0-9]+)", "$_{\\1}$", str(self.endgroupPairs))))
        plt.xlabel("m/z")
        plt.ylabel("Rel. abundance")
        if mzRange is not None:
            plt.xlim(mz for mz in mzRange)
        plt.ylim(0, 1.05)
        plt.show()

    def getIsobaricSpecies(self):
        if self.isobaricSpecies is None:
            return None
        else:
            isobaricSpecies = []
            for macromolecule in self.isobaricSpecies:
                isobaricSpecies.append([macromolecule["moleculeData"].formula_ion, macromolecule["moleculeData"].massMostAbundant, list([str(list(repUnitsCombination)), self.endgroupPairs[endgroupPair]] for repUnitsCombination, endgroupPair in zip(macromolecule["repUnitsCombinations"],macromolecule["endgroupPairIndices"]))])
            return sorted(isobaricSpecies, key=itemgetter(1))

    def printIsobaricSpecies(self):
        isobaricSpecies = self.getIsobaricSpecies()
        if isobaricSpecies is not None:
            print("Formula (ion)\tMost abundant mass\tMonomer combinations\tEnd-groups")
            print("\t" + str(self.monomers))
            for macromolecule in isobaricSpecies:
                i = 1
                for repUnitsCombination, endgroupPair in macromolecule[2]:
                    if i == 1:
                        print(macromolecule[0] + "\t" + "%.4f" % round(macromolecule[1], 4) + "\t" + repUnitsCombination + "\t" + endgroupPair)
                    else:
                        print("\t\t" + repUnitsCombination + "\t" + endgroupPair)
                    i += 1
        else:
            print("No isobaric macromolecules in the simulation...")



# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Kevin De Bruycker and Tim Krappitz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import absolute_import
from collections import Counter
import itertools
import pymacroms
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from operator import itemgetter, attrgetter
import re
from sklearn import linear_model
import sys
import os
import progressbar
from svglib.svglib import svg2rlg
# from reportlab.graphics import renderPDF
from reportlab.platypus import SimpleDocTemplate, Table, TableStyle, Paragraph, Frame, Spacer, Image, PageBreak
from reportlab.lib import colors
# from reportlab.lib.pagesizes import A3, A4, landscape, portrait
from reportlab.lib.units import cm, inch
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.lib.utils import ImageReader
from io import BytesIO

reportStyles = getSampleStyleSheet()

class Spectrum:

    def __init__(self,
                 MSData,
                 mzRange,
                 minRelAbundance = 0.01,
                 ppmDev = 5,
                 resolution = None,
                 activeScan = None,
                 mergePeaksWithinResolution = False,
                 filterByMonomer = None,
                 customMonomersDatabase = None):
        """

        :type filename: str
        :type mzRange: list
        :type minRelAbundance: float
        :type ppmDev: float
        :type resolution: float
        :type activeScan: str
        :type mergePeaksWithinResolution: bool
        :type filterByMonomer: Counter or str
        :type customMonomersDatabase: dict
        """

        self.MSData = MSData # necessary to be able to write the results later on
        self.mzRange = mzRange
        self.minRelAbundance = minRelAbundance
        self.ppmDev = ppmDev
        self.resolution = resolution
        self.filename = MSData.filename
        self.extension = MSData.extension
        if len(self.MSData.scans) == 1:
            self.activeScan = '1'
        elif activeScan is None:
            sys.exit("Multiple scans found in imported spectrum, please provide arg activeScan to Spectrum to select the one to be analysed.")
        else:
            self.activeScan = activeScan
        if customMonomersDatabase is not None:
            self.customMonomersDatabase = customMonomersDatabase
        else:
            self.customMonomersDatabase = dict()

        self.peakData = MSData.scans[self.activeScan]["spectrum"].copy(deep=True)
        self.peakData.columns = ["mz", "relAbundance"]
        # max_val = pd.max(spectrum["abundance"])
        # spectrum["relAbundance"] = spectrum.abundance/spectrum.loc[(spectrum["mz"] >= mzRange[0]) & (spectrum["mz"] <= mzRange[1])].abundance.max()
        # spectrum["outsideROI"] = ~((spectrum.mz >= mzRange[0]) & (spectrum.mz <= mzRange[1])) # condition is true when inside the range, ~ negates the condition
        # outsideROI is not necessary because the filter for repeating units will ignore the mass peaks outside the ROI and set those rows for deletion anyway
        # calculate the RelAbundance based on the abundances in the ROI only
        self.peakData.relAbundance = self.peakData.relAbundance / self.peakData.loc[(self.peakData.mz >= mzRange[0]) & (self.peakData.mz <= mzRange[1])].relAbundance.max()  # devides the abundance by the maximum of the abundances of which the mass is not outside the region of interest
        self.rawSpectrum = pd.DataFrame(self.peakData.loc[(self.peakData.mz >= mzRange[0]) & (self.peakData.mz <= mzRange[1])])
        self.rawSpectrum.reset_index(drop=True, inplace=True)

        print("Filtering spectrum...")
        if filterByMonomer is not None:
            if type(filterByMonomer) == str:
                if filterByMonomer in self.customMonomersDatabase:
                    filterByMonomer = pymacroms.getMonoIsotopicMass(self.customMonomersDatabase[filterByMonomer])
                elif filterByMonomer in pymacroms.database.monomers:
                    filterByMonomer = pymacroms.getMonoIsotopicMass(pymacroms.database.monomers[filterByMonomer])
                else:
                    sys.exit("Filter by monomer: " + filterByMonomer + " not found in either custom or built-in database!")
            else:
                filterByMonomer = pymacroms.getMonoIsotopicMass(filterByMonomer)
            # filter the spectrum for peaks that only have a corresponding peak at + and - 1 repeating unit (the first monomer is used for this assessment)
            noPlusMinOneRU = []
            for mz in self.peakData.mz:
                if mz >= mzRange[0] and mz <= mzRange[1]:
                    noPlusMinOneRU.append(not self.hasPlusMinOneRU(mz, filterByMonomer))
                else:
                    noPlusMinOneRU.append(True)
            self.peakData["noPlusMinOneRU"] = noPlusMinOneRU
            print("--> Peaks outside mass range removed")
            print("--> Peaks without a +/- 1 repeating unit counterpart removed")
        else:
            self.peakData["noPlusMinOneRU"] = ~((self.peakData.mz >= mzRange[0]) & (self.peakData.mz <= mzRange[1]))  # condition is true when inside the range, ~ negates the condition
            print("--> Peaks outside mass range removed")
        self.peakData.drop(self.peakData[self.peakData["noPlusMinOneRU"]].index, inplace=True)
        self.peakData.drop(columns=["noPlusMinOneRU"], inplace=True)
        self.peakData.reset_index(drop=True, inplace=True)
        if mergePeaksWithinResolution:
            self.peakData = pymacroms.combineIsotopes(self.peakData, resolution)
            self.peakData.relAbundance = self.peakData.relAbundance / self.peakData.relAbundance.max() # recalculate relative abundance
            print("--> Merged peaks that are not separated with a resolution of " + str(self.resolution))
        # tag the entries that have a too low RelAbund to be deleted
        # self.peakData.loc[(self.peakData.relAbundance < minRelAbundance)].relAbundance = None # Doesn't work, only works when setting the whole 'matched' row to None, but mz values still needed for later matching of +- 1 RU
        self.peakData["tooLowRelAbund"] = self.peakData.relAbundance < minRelAbundance
        # self.peakData["toBeDeleted"] = self.peakData.tooLowRelAbund | self.peakData.noPlusMinOneRU
        self.peakData.drop(self.peakData[self.peakData["tooLowRelAbund"]].index, inplace=True)
        self.peakData.drop(columns=["tooLowRelAbund"], inplace=True)
        self.peakData.reset_index(drop=True, inplace=True)
        print("--> Peaks with too low relative abundance removed")
        print("--> Done\n")

    def hasPlusMinOneRU(self, mz, RUMass):
        mzRangeMinOne = [mz - mz * self.ppmDev * 1e-6 - RUMass - (mz - mz * self.ppmDev * 1e-6 - RUMass) * self.ppmDev * 1e-6,
                         mz + mz * self.ppmDev * 1e-6 - RUMass + (mz + mz * self.ppmDev * 1e-6 - RUMass) * self.ppmDev * 1e-6]
        # The 'true' mass of an experimental peak should be within a certain ppm range of that experimental peak,
        # but the experimental peak can again be ppmDev apart from that range -RUMass
        mzRangePlusOne = [mz - mz * self.ppmDev * 1e-6 + RUMass - (mz - mz * self.ppmDev * 1e-6 + RUMass) * self.ppmDev * 1e-6,
                          mz + mz * self.ppmDev * 1e-6 + RUMass + (mz + mz * self.ppmDev * 1e-6 + RUMass) * self.ppmDev * 1e-6]
        if len(self.peakData.loc[(self.peakData.mz >= mzRangeMinOne[0]) & (self.peakData.mz <= mzRangeMinOne[1])]) > 0 and len(
                self.peakData.loc[(self.peakData.mz >= mzRangePlusOne[0]) & (self.peakData.mz <= mzRangePlusOne[1])]) > 0:
            return True
        else:
            return False

    def matchPeaks(self, polymerSimulation):
        self.polymerSimulation = polymerSimulation

        print("Matching experimental spectrum to peaks of the simulated polymer...")
        print("(accuracy = " + str(self.ppmDev) + " ppm, resolution = " + str(self.resolution) + ")")
        if self.resolution is not None: # check if the isotopes have to be aggregated
            # check if the isotopes have been aggregated yet for the first macromolecule,
            # if yes then it should have been done for all of them and this step can be skipped
            if hasattr(self.polymerSimulation.macromolecules[0]["moleculeData"], 'resolution'):
                if self.polymerSimulation.macromolecules[0]["moleculeData"].resolution == self.resolution:
                    aggregateIsotopes = False
                else:
                    aggregateIsotopes = True
            else:
                aggregateIsotopes = True
        else:
            aggregateIsotopes = False
        if aggregateIsotopes:
            self.polymerSimulation.aggregateIsotopes(self.resolution)
        print("--> Total amount of experimental peaks: " + str(len(self.peakData)))
        theoPeakContributions = []  # for every experimental peak, this variable contains the theoretical composition:
        # [[indices of contributing macromolecules], [corresponding normalised abundances]]
        # identifiedMacromolecules = []
        matchedIsotopes = dict()
        peakMatched = []
        print("--> Matching...")
        sys.stdout.flush()
        for mz, relAbundance in progressbar.progressbar(self.peakData.itertuples(index=False, name=None), 0, len(self.peakData)):
            moleculeList = self.polymerSimulation.getMoleculeList(mz, self.ppmDev, self.resolution)
            if moleculeList is not None:
                macromolecules = []
                normAbundances = []
                for macromoleculeIndex, repUnitsCombination, isotopesNearMass in moleculeList:
                    macromolecules.append(macromoleculeIndex)
                    normAbundances.append(sum(list(zip(*isotopesNearMass))[1])) # the normAbundance, i.e. contribution of 1 macromolecule to the experimental peak is the sum of the normalised abundances within the accuracy of ppmDev
                    if macromoleculeIndex in matchedIsotopes:
                        matchedIsotopes[macromoleculeIndex] += list((mz, isotopeMass, deviation) for isotopeMass, normAbundance, deviation in isotopesNearMass)
                    else:
                        matchedIsotopes[macromoleculeIndex] = list((mz, isotopeMass, deviation) for isotopeMass, normAbundance, deviation in isotopesNearMass)
                theoPeakContributions.append([macromolecules, normAbundances])
                # identifiedMacromolecules += macromolecules
                # matchedCalcMass.append(np.average(masses, weights=massWeightingFactors))
                peakMatched.append(True)
            else:
                # matchedCalcMass.append(None)
                peakMatched.append(False)
        self.peakData["peakMatched"] = peakMatched
        self.theoPeakContributions = theoPeakContributions
        # self.identifiedMacromolecules = np.unique(identifiedMacromolecules)
        # self.identifiedMacromolecules = sorted(list(matchedIsotopes.keys()))
        self.matchedIsotopes = matchedIsotopes
        print("--> %i peaks successfully matched" % len(self.peakData[self.peakData.peakMatched]))
        print("--> These peaks are (potentially) originating from %i macromolecules" % len(self.matchedIsotopes))
        print("--> %i experimental m/z values could not be matched to a calculated m/z" % len(self.peakData[~self.peakData.peakMatched]))
        print("--> Done\n")

    def plotMatchPeaksOverview(self, mzRange: list = None, showMatched: bool = True, showNotMatched: bool = True, savePlotAs: str = None, reportOutput: bool = False):
        y_val = 0 if reportOutput else -0.1
        if showMatched:
            mz_axis = []
            relAbund_axis = []
            for mz, relAbund in self.peakData[["mz", "relAbundance"]].loc[self.peakData.peakMatched].itertuples(index=False, name=None):
                mz_axis += list([mz - 1e-10, mz, mz + 1e-10])
                relAbund_axis += list([y_val, relAbund, y_val])
            plt.plot(mz_axis, relAbund_axis, color="green", label="Matched to theoretical m/z")
        if showNotMatched:
            mz_axis = []
            relAbund_axis = []
            for mz, relAbund in self.peakData[["mz", "relAbundance"]].loc[~self.peakData.peakMatched].itertuples(index=False, name=None):
                mz_axis += list([mz - 1e-10, mz, mz + 1e-10])
                relAbund_axis += list([y_val, relAbund, y_val])
            plt.plot(mz_axis, relAbund_axis, color="red", label="Not matched to theoretical m/z")
        plt.legend()
        plt.title("Filtered experimental spectrum")
        plt.xlabel("m/z")
        plt.ylabel("Rel. abundance")
        if mzRange is not None:
            plt.xlim(mz for mz in mzRange)
        plt.ylim(0, 1.05)
        if savePlotAs is not None:
            plt.savefig(self.filename + "_matchPeaksOverview." + savePlotAs, dpi=300, format=savePlotAs)
        if reportOutput:
            imgdata = BytesIO()
            plt.savefig(imgdata, format="svg")
            plt.close()
            imgdata.seek(0)  # rewind the data
            return Image(svg2rlg(imgdata))
        else:
            plt.show()
            # plt.close()

    def calcComposition(self, minimumAmount: float = 0, minimumIsotopeFraction: float = 0.1):
        self.compositionMinimumAmount = minimumAmount # save for future reference
        print("Calculating the composition by fitting the experimental and theoretical intensities...")
        # target = pd.DataFrame(np.zeros(len(self.theoPeakComposition)), columns=["relAbundance"])
        target = pd.DataFrame(self.peakData.loc[self.peakData.peakMatched].relAbundance, columns=["relAbundance"])
        # target.reset_index(drop=True, inplace=True)
        df = pd.DataFrame(np.zeros((len(self.theoPeakContributions), len(self.matchedIsotopes))), columns=sorted(list(self.matchedIsotopes.keys())))
        for index, [macromolecules, normAbundances] in enumerate(self.theoPeakContributions): #index, [indices of contributing macromolecules], [corresponding normalised abundances]]
            # target.iloc[index].relAbundance = relAbundance
            for macromolecule, normAbundance in zip(macromolecules, normAbundances):
                df.loc[index, macromolecule] = normAbundance
        df.index = target.index
        dfColumnsSum = pd.DataFrame(df.sum(), columns=["dfSum"])
        # this sum represents the fraction of the (normAbundance of the) isotopic pattern that is matched in the spectrum:
        # if all the peaks of the theoretical pattern are matched, this sum should be 1;
        # if only 1 minor peak of the isotopic pattern is found in the experimental spectrum, the sum will be low --> probably mismatch
        df.drop(columns=dfColumnsSum.loc[dfColumnsSum.dfSum < minimumIsotopeFraction].index, inplace=True)
        regression = linear_model.LinearRegression(fit_intercept=False)
        regression.fit(df, target)
        regressionR2 = regression.score(df, target)
        print("--> Composition found with an R^2 of %.5f" % round(regressionR2, 5))
        regResult = pd.DataFrame(np.transpose(regression.coef_), index=df.columns, columns=["amounts"])
        # self.peakData["fittedRelAbundance"] = pd.DataFrame(np.dot(df, regResult), index=target.index, columns=["calcRelAbundance"]).calcRelAbundance # Don't do this, use the dedicated method getFittedIsotopicDist instead: this will exclude isotopes that were not matched while they will be included when calculating the whole spectrum
        self.peakData["diffRelAbundance"] = pd.DataFrame(np.dot(df, regResult), index=target.index, columns=["calcRelAbundance"]).calcRelAbundance - target.relAbundance
        regResult.amounts = regResult.amounts / regResult.amounts.max() # normalisation
        while regResult.amounts.min() < minimumAmount:
            print("--> One or more macromolecules have a contribution\n    that is either negative or below the threshold of %.2f!" % round(minimumAmount, 2))
            print("--> Recalculating...")
            # remove values that are too low and repeat linear regression
            df.drop(columns=regResult.loc[regResult.amounts < minimumAmount].index, inplace=True)
            regression.fit(df, target)
            regressionR2 = regression.score(df, target)
            print("--> Composition found with an R^2 of %.5f" % round(regressionR2, 5))
            regResult = pd.DataFrame(np.transpose(regression.coef_), index=df.columns, columns=["amounts"])
            self.peakData["diffRelAbundance"] = pd.DataFrame(np.dot(df, regResult), index=target.index, columns=["calcRelAbundance"]).calcRelAbundance - target.relAbundance
            regResult.amounts = regResult.amounts / regResult.amounts.max() # normalisation
        if len(regResult) == 0:
            self.composition = None
            self.regressionR2 = None
        else:
            self.composition = regResult
            self.regressionR2 = regressionR2
            composition_processed = []
            for macromolecule, amount in regResult.itertuples(name=None):
                composition_processed.append([list(self.polymerSimulation.macromolecules[macromolecule]["repUnitsCombination"]),
                                    self.polymerSimulation.endgroupPairs[
                                        self.polymerSimulation.macromolecules[macromolecule]["endgroupPairIndex"]],
                                    self.polymerSimulation.macromolecules[macromolecule]["moleculeData"].formula_ion,
                                    amount,
                                    self.polymerSimulation.macromolecules[macromolecule]["moleculeData"].molecularWeight,
                                    self.polymerSimulation.macromolecules[macromolecule][
                                        "moleculeData"].massMostAbundant,
                                    self.matchedIsotopes[macromolecule]])
            self.MSData.scans[self.activeScan]['composition'] = sorted(composition_processed, key=itemgetter(4))
            self.MSData.scans[self.activeScan]['Mn'] = sum(list(Mw * N for a, b, c, N, Mw, d, e in composition_processed)) / sum(list(N for a, b, c, N, Mw, d, e in composition_processed))
            self.MSData.scans[self.activeScan]['Mw'] = sum(list(Mw * Mw * N for a, b, c, N, Mw, d, e in composition_processed)) / sum(list(Mw * N for a, b, c, N, Mw, d, e in composition_processed))
            self.MSData.scans[self.activeScan]['dispersity'] = self.MSData.scans[self.activeScan]['Mw'] / self.MSData.scans[self.activeScan]['Mn']
        print("--> Done\n")

    def getComposition(self):
        composition = []
        for macromolecule, amount in self.composition.itertuples(name=None):
            composition.append([list(self.polymerSimulation.macromolecules[macromolecule]["repUnitsCombination"]),
                           self.polymerSimulation.endgroupPairs[
                               self.polymerSimulation.macromolecules[macromolecule]["endgroupPairIndex"]],
                           self.polymerSimulation.macromolecules[macromolecule]["moleculeData"].formula_ion,
                           amount,
                           self.polymerSimulation.macromolecules[macromolecule]["moleculeData"].massMostAbundant,
                           self.matchedIsotopes[macromolecule]])
        return sorted(composition, key=itemgetter(4))

    def printComposition(self):
        if not hasattr(self, "composition"):
            sys.exit("Calculate the composition first using Spectrum.calcComposition")
        if self.composition is not None:
            composition = self.getComposition()
            print("R^2 = %.5f" % round(self.regressionR2, 5))
            print("Monomer combination\tEnd-groups\tFormula (ion)\tAmount\tMost abundant mass\tMatched isotopes")
            print(str(self.polymerSimulation.monomers) + "\t\t\t\t\t(m/z (exp + offset of " + str(self.MSData.mzOffset)+ "), m/z (calc), dev (ppm))")
            for repUnitsCombination, endgroupPair, formula_ion, amount, massMostAbundant, matchedIsotopes in composition:
                print(str(repUnitsCombination) + "\t" + endgroupPair + "\t" + formula_ion + "\t" + "%.2f" % round(amount, 2) + "\t" + str(massMostAbundant) + "\t" + str(matchedIsotopes))
        else:
            print("No matching molecules found...")

    def saveComposition(self):
        if not hasattr(self, "composition"):
            sys.exit("Calculate the composition first using Spectrum.calcComposition")
        if self.composition is not None:
            composition = self.getComposition()
            export = pd.DataFrame(list((repUnitsCombination, endgroupPair, formula_ion, amount, massMostAbundant, list(matchedIsotopes)) for repUnitsCombination, endgroupPair, formula_ion, amount, massMostAbundant, matchedIsotopes in composition), columns=["Monomer combination", "End-groups", "Formula (ion)", "Amount", "Most abundant mass", "Matched isotopes (m/z (exp + offset of " + str(self.MSData.mzOffset)+ "), m/z (calc), dev (ppm))"])
        else:
            export = pd.DataFrame(columns=["No matching molecules found..."])
        if len(self.MSData.scans) == 1:
            export.to_csv(self.filename + "_composition.csv", index=False)
        else:
            export.to_csv(self.filename + "_composition_scan" + self.activeScan + ".csv", index=False)


        isobaricSpecies = self.polymerSimulation.getIsobaricSpecies()
        if isobaricSpecies is not None:
            report = []
            for macromolecule in isobaricSpecies:
                i = 1
                for repUnitsCombination, endgroupPair in macromolecule[2]:
                    if i == 1:
                        report.append([macromolecule[0], macromolecule[1], repUnitsCombination, endgroupPair])
                    else:
                        report.append([None, None, repUnitsCombination, endgroupPair])
                    i += 1
            export = pd.DataFrame(list((formula_ion, massMostAbundant, repUnitsCombination, endgroupPair) for formula_ion, massMostAbundant, repUnitsCombination, endgroupPair in report),
                                  columns=["Formula (ion)", "Most abundant mass", "Monomer combinations", "End-groups"])
        else:
            export = pd.DataFrame(columns=["No isobaric macromolecules in the simulation..."])
        if len(self.MSData.scans) == 1:
            export.to_csv(self.filename + "_isobaric.csv", index=False)
        else:
            export.to_csv(self.filename + "_isobaric_scan" + self.activeScan + ".csv", index=False)

    def getFittedIsotopicDist(self, resolution: float = None):
        if not hasattr(self, "composition"):
            sys.exit("Calculate the composition first using Spectrum.calcComposition")
        if self.composition is not None:
            fittedIsotopicDist = pd.DataFrame(columns=["mz", "normAbundance"])
            for macromolecule, amount in self.composition.itertuples(name=None):
                for mz, normAbund in self.polymerSimulation.macromolecules[macromolecule]["moleculeData"].isotopicDist:
                    fittedIsotopicDist = fittedIsotopicDist.append(pd.DataFrame([[mz, normAbund * amount]], columns=["mz", "normAbundance"]), ignore_index=True)
            fittedIsotopicDist.normAbundance = fittedIsotopicDist.normAbundance / fittedIsotopicDist.normAbundance.sum() # normalise
            # fittedIsotopicDist["relAbundance"] = fittedIsotopicDist.normAbundance / fittedIsotopicDist.normAbundance.max()
            fittedIsotopicDist.sort_values("mz", inplace=True)
            fittedIsotopicDist.reset_index(drop=True, inplace=True)
            if resolution is not None:
                return pymacroms.combineIsotopes(fittedIsotopicDist, resolution)
            else:
                return fittedIsotopicDist
        else:
            return None

    def printFittedSpectrum(self, resolution: float = None):
        fittedIsotopicDist = self.getFittedIsotopicDist(resolution)
        if fittedIsotopicDist is not None:
            print("Fitted spectrum")
            print("Mass\tRel. Abundance")
            for mz, relAbund in pymacroms.toRelativeAbundance(list(fittedIsotopicDist.itertuples(index=False, name=None))):
                print(str(mz) + "\t" + str(relAbund))
        else:
            print("No matching molecules found...")

    def plotFittedSpectrum(self, resolution: float = None, color = "black", mzRange: list = None):
        fittedIsotopicDist = self.getFittedIsotopicDist(resolution)
        if fittedIsotopicDist is not None:
            mz_axis = []
            relAbund_axis = []
            for mz, relAbund in pymacroms.toRelativeAbundance(list(fittedIsotopicDist.itertuples(index=False, name=None))):
                mz_axis += list([mz - 1e-10, mz, mz + 1e-10])
                relAbund_axis += list([-0.1, relAbund, -0.1])
            plt.plot(mz_axis, relAbund_axis, color=color)
            plt.title("Fitted spectrum")
            plt.xlabel("m/z")
            plt.ylabel("Rel. abundance")
            if mzRange is not None:
                plt.xlim(mz for mz in mzRange)
            plt.ylim(0, 1.05)
            plt.show()
        else:
            print("No matching molecules found...")

    def plotComparisonSpectrum(self, resolution: float = None, color = "black", mzRange: list = None, savePlotAs: str = None, reportOutput: bool = False):
        fittedIsotopicDist = self.getFittedIsotopicDist(resolution)
        if fittedIsotopicDist is not None:
            y_val = 0 if reportOutput else -0.1
            rawSpectrum = plt.subplot(211)
            mz_axis = []
            relAbund_axis = []
            for mz, relAbund in self.rawSpectrum.itertuples(index=False, name=None):
                mz_axis += list([mz - 1e-10, mz, mz + 1e-10])
                relAbund_axis += list([y_val, relAbund, y_val])
            rawSpectrum.plot(mz_axis, relAbund_axis, color=color, label="Raw data")
            plt.legend()
            plt.title("Spectrum comparison")
            plt.ylabel("Rel. abundance")
            if mzRange is not None:
                plt.xlim(mz for mz in mzRange)
            plt.ylim(0, 1.05)
            fittedSpectrum = plt.subplot(212, sharex=rawSpectrum, sharey=rawSpectrum)
            mz_axis = []
            relAbund_axis = []
            for mz, relAbund in pymacroms.toRelativeAbundance(list(fittedIsotopicDist.itertuples(index=False, name=None))):
                mz_axis += list([mz - 1e-10, mz, mz + 1e-10])
                relAbund_axis += list([y_val, relAbund, y_val])
            fittedSpectrum.plot(mz_axis, relAbund_axis, color=color, label="Fitted spectrum")
            # plt.title(r"Isotopic distribution ({}{} adducts) of a (co)polymer of".format(self.amountAdductIons,
            #                                                                              re.sub("([0-9]+)", "$_{\\1}$",
            #                                                                                     re.sub("([+-]+)",
            #                                                                                            "$^{\\1}$",
            #                                                                                            self.adductIon))) + "\n" +
            #           "{} with end-group(s): {}".format(re.sub("([0-9]+)", "$_{\\1}$", str(self.monomers)),
            #                                             re.sub("([0-9]+)", "$_{\\1}$", str(self.endgroupPairs))))
            plt.legend()
            plt.xlabel("m/z")
            plt.ylabel("Rel. abundance")
            if mzRange is not None:
                plt.xlim(mz for mz in mzRange)
            plt.ylim(0, 1.05)
            if savePlotAs is not None:
                plt.savefig(self.filename + "_comparisonSpectrum." + savePlotAs, dpi=300, format=savePlotAs)
            if reportOutput:
                imgdata = BytesIO()
                plt.savefig(imgdata, format="svg")
                plt.close()
                imgdata.seek(0)  # rewind the data
                return Image(svg2rlg(imgdata))
            else:
                plt.show()
                # plt.close()
        else:
            if reportOutput:
                return Paragraph("No isobaric macromolecules in the simulation...", reportStyles['Normal'])
            else:
                print("No matching molecules found...")

    def plotDiffIsotopicDist(self, mzRange: list = None, showExperimental: bool = True, showNonMatched: bool = False, showFitted: bool = False, resolution: float = None):
        if not hasattr(self, "composition"):
            sys.exit("Calculate the composition first using Spectrum.calcComposition")
        if self.composition is not None:
            filteredSpectrum = plt.subplot(211)
            if showExperimental:
                mz_axis = []
                relAbund_axis = []
                for mz, relAbund in self.peakData[["mz", "relAbundance"]].loc[self.peakData.peakMatched].itertuples(
                        index=False, name=None):
                    mz_axis += list([mz - 1e-10, mz, mz + 1e-10])
                    relAbund_axis += list([-0.1, relAbund, -0.1])
                filteredSpectrum.plot(mz_axis, relAbund_axis, color="green", label="(Filtered) matched experimental")
            if showNonMatched:
                mz_axis = []
                relAbund_axis = []
                for mz, relAbund in self.peakData[["mz", "relAbundance"]].loc[~self.peakData.peakMatched].itertuples(
                        index=False, name=None):
                    mz_axis += list([mz - 1e-10, mz, mz + 1e-10])
                    relAbund_axis += list([-0.1, relAbund, -0.1])
                filteredSpectrum.plot(mz_axis, relAbund_axis, color="red", label="Not matched to theoretical m/z")
            if showFitted:
                fittedIsotopicDist = self.getFittedIsotopicDist(resolution)
                if fittedIsotopicDist is not None:
                    mz_axis = []
                    relAbund_axis = []
                    for mz, relAbund in pymacroms.toRelativeAbundance(list(fittedIsotopicDist.itertuples(index=False, name=None))):
                        mz_axis += list([mz - 1e-10, mz, mz + 1e-10])
                        relAbund_axis += list([-0.1, relAbund, -0.1])
                    filteredSpectrum.plot(mz_axis, relAbund_axis, color="blue", label="(Fitted) calculated")
            plt.legend()
            plt.title("Filtered experimental spectrum")
            # plt.xlabel("m/z")
            plt.ylabel("Rel. abundance")
            if mzRange is not None:
                plt.xlim(mz for mz in mzRange)
            plt.ylim(0, 1.05)

            diffFittedSpectrum = plt.subplot(212, sharex=filteredSpectrum)
            mz_axis = []
            diffRelAbund_axis = []
            for mz, diffRelAbund in self.peakData[["mz", "diffRelAbundance"]].loc[self.peakData.peakMatched].itertuples(index=False, name=None):
                mz_axis += list([mz - 1e-10, mz, mz + 1e-10])
                diffRelAbund_axis += list([0, diffRelAbund, 0])
            diffFittedSpectrum.plot(mz_axis, diffRelAbund_axis, color="black", label="Difference fitted-experimental")
            plt.legend()
            plt.xlabel("m/z")
            plt.ylabel("Rel. abundance")
            if mzRange is not None:
                plt.xlim(mz for mz in mzRange)
            # plt.ylim(-0.55, 0.55)
            # plt.savefig("output.pdf", dpi=300, format='pdf')
            plt.show()
            # plt.close()

    def saveReport(self, plotMatchPeaksOverview=True, plotComparisonSpectrum=True, composition=True, savePlotsAs: str = None, AppMolFr=False):
        if len(self.MSData.scans) == 1:
            reportFile = SimpleDocTemplate(self.filename + "_report.pdf", rightMargin=0.5 * inch, leftMargin=0.5 * inch, topMargin=0.5 * inch, bottomMargin=0.5 * inch)
        else:
            reportFile = SimpleDocTemplate(self.filename + "_report_scan" + self.activeScan + ".pdf", rightMargin=0.5 * inch, leftMargin=0.5 * inch, topMargin=0.5 * inch, bottomMargin=0.5 * inch)
        reportContent = []

        reportContent.append(Paragraph("Experiment " + re.sub("\A(.*)/([^/]*)\Z", "\\2", self.filename), reportStyles['Title']))

        reportContent.append(Paragraph("Processing parameters", reportStyles['Heading1']))
        parametersTable = [["mzRange", "mzOffset", "ppmDev", "resolution", "minRelAbundance"]]
        parametersTable.append([str(self.mzRange), self.MSData.mzOffset, self.ppmDev,
                          self.resolution, self.minRelAbundance])
        parametersTable.append([])
        table = Table(parametersTable, repeatRows=2, style=[
            ('BACKGROUND', (0, 0), (-1, 0), colors.Color(0, 0.7, 0.7)),
            ('LINEBELOW', (0, 0), (-1, 0), 1, colors.black),
            ('ALIGN', (0, 0), (-1, -1), 'CENTER'),
        ])
        reportContent.append(table)

        parametersTable = [["endgroupPairs", "monomers", "adductIon", "charge"]]
        parametersTable.append(
            [self.polymerSimulation.endgroupPairs[0], self.polymerSimulation.monomers[0],
             self.polymerSimulation.adductIon, self.polymerSimulation.charge])
        for endgroup, monomer in itertools.zip_longest(self.polymerSimulation.endgroupPairs[1:], self.polymerSimulation.monomers[1:], fillvalue=''):
            parametersTable.append([endgroup, monomer])
        table = Table(parametersTable, repeatRows=2, style=[
            ('BACKGROUND', (0, 0), (-1, 0), colors.Color(0, 0.7, 0.7)),
            ('LINEBELOW', (0, 0), (-1, 0), 1, colors.black),
            ('ALIGN', (0, 0), (-1, -1), 'CENTER'),
        ])
        reportContent.append(table)
        reportContent.append(Spacer(1 * cm, 1 * cm))


        # reportContent.append(PageBreak())


        if plotMatchPeaksOverview:
            elementTitle = Paragraph("Matching overview", reportStyles['Heading1'])
            elementTitle.keepWithNext = True
            reportContent.append(elementTitle)
            reportContent.append(self.plotMatchPeaksOverview(reportOutput=True, savePlotAs=savePlotsAs))

        if plotComparisonSpectrum:
            elementTitle = Paragraph("Experimental vs. calculated spectrum", reportStyles['Heading1'])
            elementTitle.keepWithNext = True
            reportContent.append(elementTitle)
            reportContent.append(self.plotComparisonSpectrum(resolution=self.resolution, color="darkblue", reportOutput=True, savePlotAs=savePlotsAs))

        #
        # if AppMolFr:  # dangerous part, if this is called before the composition table it breaks the code
        #     story.append(pymacroms.apparentMoleFractions(
        #         filename=filename,
        #         charge=polymerSimulation.charge,
        #         adductIon=polymerSimulation.adductIon,
        #         endgroupPairs=polymerSimulation.endgroupPairs,
        #         monomers=polymerSimulation.monomers,
        #         minRepeatingUnits=10,
        #         maxRepeatingUnits=25,
        #         ppmDev=spectrumExperimental.ppmDev,
        #     ).plotApparentMoleFractions())

        if composition:
            reportContent.append(Spacer(1 * cm, 1 * cm))
            reportContent.append(Paragraph("Identified species", reportStyles['Heading1']))
            composition = self.getComposition()

            compositionTable = [["Polymer composition", "", "", "", "", "Matched peaks", "", ""],
                    ["Monomers", "Endgroups", "Formula & ion", "Amount", "Most abundant\nisotope", "m/z\n(exp+offset)",
                     "m/z\n(calc)", "\u0394\n(ppm)"]]
            for repUnitsCombination, endgroupPair, formula_ion, amount, massMostAbundant, matchedIsotopes in composition:
                compositionTable.append([str(repUnitsCombination), endgroupPair, formula_ion,
                                         "%.2f" % round(amount, 2), "%.4f" % round(massMostAbundant, 4),
                                         "%.4f" % round(matchedIsotopes[0][0], 4), "%.4f" % round(matchedIsotopes[0][1], 4), "%.2f" % round(matchedIsotopes[0][2], 2)])
                for mz_exp, mz_calc, mz_dev in matchedIsotopes[1:]:  # skip the first tuple, that is dealt with in the previous loop
                    compositionTable.append(
                        ["", "", "", "", "", "%.4f" % round(mz_exp, 4), "%.4f" % round(mz_calc, 4), "%.2f" % round(mz_dev, 2)])
            table = Table(compositionTable, repeatRows=2, style=[
                ('BACKGROUND', (0, 0), (-1, 1), colors.Color(0, 0.7, 0.7)),
                ('SPAN', (0, 0), (1, 0)),
                ('SPAN', (-3, 0), (-1, 0)),
                ('LINEBELOW', (0, 1), (-1, 1), 1, colors.black),
                ('ALIGN', (0, 0), (-1, -1), 'CENTER'),
            ])
            reportContent.append(table)

            reportContent.append(Spacer(1 * cm, 1 * cm))
            reportContent.append(Paragraph("Isobaric species", reportStyles['Heading1']))
            isobaricSpecies = self.polymerSimulation.getIsobaricSpecies()
            if isobaricSpecies is not None:
                isobaricTable = [["Formula (ion)", "Most abundant mass", "Monomer combinations", "End-groups"]]
                for macromolecule in isobaricSpecies:
                    isobaricTable.append([macromolecule[0], macromolecule[1], macromolecule[2][0][0], macromolecule[2][0][1]])
                    for repUnitsCombination, endgroupPair in macromolecule[2][1:]:
                        isobaricTable.append([None, None, repUnitsCombination, endgroupPair])
                table = Table(isobaricTable, repeatRows=2, style=[
                    ('BACKGROUND', (0, 0), (-1, 0), colors.Color(0, 0.7, 0.7)),
                    ('LINEBELOW', (0, 0), (-1, 0), 1, colors.black),
                    ('ALIGN', (0, 0), (-1, -1), 'CENTER'),
                ])
                reportContent.append(table)
            else:
                reportContent.append(Paragraph("No isobaric macromolecules in the simulation...", reportStyles['Normal']))


        reportFile.build(reportContent)









# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Kevin De Bruycker and Tim Krappitz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import absolute_import
from math import exp, log, sqrt
from collections import Counter
import itertools
import pymacroms
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from operator import itemgetter, attrgetter
import re
import sys
# from numba import jit

def getMonoIsotopicMass(formula: Counter):
    return sum(list(sorted(pymacroms.database.isotopic_abundances[element], key=itemgetter(1), reverse=True)[0][0]*amount for element, amount in formula.items()))
    # totalMass = 0
    # for element, amount in formula.items():
    #     totalMass += pymacroms.database.isotopic_abundances[element][0][0] * amount
    # return totalMass

def getMolecularWeight(formula: Counter):
    return sum(list(pymacroms.database.atomic_weights[element]*amount for element, amount in formula.items()))


def getCounterFromFormula(formula: str):
    return Counter({element: int(amount) for element, amount in re.findall("([A-Z][a-z]*)([0-9]+)", re.sub(" ", "", re.sub("  ", " 1 ", re.sub("([A-Z][a-z]*)([0-9]*)", "\\1 \\2 ", re.sub(" ","", formula)))))})
    #first removes all spaces, then replace all elements with optional number by element-space-amount-space
    #because an absent number will result in a double space, the double spaces are then replaced by space-1-space
    #by removing the spaces again, the formula is then guaranteed to have an amount for every element (C9H18NO --> C9H18N1O1)
    #finally, transform in a list and then Counter by matching the element-amount pairs

    '''
    return Counter({element: int(amount) for element, amount in re.findall("([A-Z][a-z]*)([0-9]*)", re.sub("([A-Z][a-z]*)([A-Z]|\Z)", "\g<1>1\g<2>", formula))})
    Is more simple but gives an error on for example C9H18NO because in the first substitution, NO is replaced by N1O, and because O is already matched, it is not replaced by O1
    This results in an O: '' in the library and thus an error when converting to int 
    '''

def getFormulaStrFromCounter(formula: Counter):
    tempFormula = ""
    for element, amount in formula.items():  # sorted(dict) does not influence the calculations, while leaving it out typically gives more OK formulas for the OCDd chemist
        tempFormula += element + str(amount)
    return tempFormula

def toRelativeAbundance(peaklist: list):
    maxAbund = max(list(zip(*peaklist))[1])
    return sorted(list((mass, normAbund/maxAbund) for mass, normAbund in peaklist))

def combineIsotopes(peaklist: list, resolution: float):
    # returns a list of peaks [(mz, normAbundance)]
    if type(peaklist) == list:
        peaklist_df = pd.DataFrame(peaklist, columns=["mz", "abundance"])
    else:
        columns_temp = peaklist.columns
        peaklist_df = peaklist
        peaklist_df.columns = ["mz", "abundance"]

    index = 0
    while index < len(peaklist_df):
        if not peaklist_df.iloc[index].isnull().all():  # checks that 'not all values in a row are NaN', alternative: .any()
            mz = peaklist_df.iloc[index].mz
            # stdDev = mz / (2 * sqrt(2 * log(2)) * resolution); +- 2stdDev contains 95% of the normal distribution while +- 1.17741stdDev is the FWHM range
            dmz = mz / (2 * resolution)
            masses = list(peaklist_df.loc[(peaklist_df.mz >= mz - dmz) & (peaklist_df.mz <= mz + dmz)].mz)
            if len(masses) > 1:
                abundances = list(peaklist_df.loc[(peaklist_df.mz >= mz - dmz) & (peaklist_df.mz <= mz + dmz)].abundance)
                peaklist_df.loc[(peaklist_df.mz >= mz - dmz) & (peaklist_df.mz <= mz + dmz)] = None
                peaklist_df.iloc[index] = np.average(masses, weights=abundances), sum(abundances)
        index += 1
    peaklist_df.dropna(how="all", inplace=True)  # get rid of all the empty lines, how=all limits to rows that are completely empty
    peaklist_df.reset_index(drop=True, inplace=True)
    # peaklist_df.normAbundance = peaklist_df.normAbundance / peaklist_df.normAbundance.sum()  # recalculate normalised abundance
    # if the sum = 1 for the input (normAbundance), then the sum will still be 1 when simply adding peaks together.
    # if the input is no normAbundance, then the recalculation has to be done externally

    if type(peaklist) == list:
        return list(peaklist_df.itertuples(index=False, name=None))
    else:
        peaklist_df.columns = columns_temp
        return peaklist_df

# Try to speed up this step using C
# # @njit
# def combineIsotopes2(peaklist: list, resolution: float):
#     # returns a list of peaks [(mz, normAbundance)]
#     if type(peaklist) == list:
#         peaklist_df = pd.DataFrame(peaklist, columns=["mz", "abundance"])
#     else:
#         columns_temp = peaklist.columns
#         peaklist_df = peaklist
#         peaklist_df.columns = ["mz", "abundance"]
#
#     index = 0
#     while not peaklist_df.iloc[index].isnull().all():
#         mz = peaklist_df.iloc[index].mz
#         range = [mz - mz / resolution / 2, mz + mz / resolution / 2]
#         masses = list(peaklist_df.loc[(peaklist_df.mz >= range[0]) & (peaklist_df.mz <= range[1])].mz)
#         if len(masses) > 1:
#             abundances = list(peaklist_df.loc[(peaklist_df.mz >= range[0]) & (peaklist_df.mz <= range[1])].abundance)
#             peaklist_df.loc[(peaklist_df.mz >= range[0]) & (peaklist_df.mz <= range[1])] = None
#             peaklist_df.iloc[index] = np.average(masses, weights=abundances), sum(abundances)
#             peaklist_df.sort_values(by=["mz"], inplace=True, na_position='last')
#         index += 1
#     peaklist_df.dropna(how="all", inplace=True)  # get rid of all the empty lines, how=all limits to rows that are completely empty
#     peaklist_df.reset_index(drop=True, inplace=True)
#     # peaklist_df.normAbundance = peaklist_df.normAbundance / peaklist_df.normAbundance.sum()  # recalculate normalised abundance
#     # if the sum = 1 for the input (normAbundance), then the sum will still be 1 when simply adding peaks together.
#     # if the input is no normAbundance, then the recalculation has to be done externally
#
#     if type(peaklist) == list:
#         return list(peaklist_df.itertuples(index=False, name=None))
#     else:
#         peaklist_df.columns = columns_temp
#         return peaklist_df








# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Kevin De Bruycker and Tim Krappitz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import absolute_import
from math import exp, isclose
from collections import Counter
import itertools
import pymacroms
import IsoSpecPy
import matplotlib.pyplot as plt
from operator import itemgetter, attrgetter
import re
import sys

class Molecule:

    def __init__(self, formula, adductIon = "H+", charge = 1, minRelAbundance = 0.01):
        """
        only the most abundant isotopes with a relative abundance of <minRelAbundance> will be taken into account

        :type formula: Counter or str
        :type adductIon: str
        :type charge: int
        :type minRelAbundance: float
        """
        #input processing: internal formula should be a counter for processing while self.formula should be a str for reporting
        if type(formula) == str:
            # save the formula as an attribute and then convert it to a counter to be able to include the ionisation
            self.formula = formula
            formula = pymacroms.getCounterFromFormula(formula)
        else:
            # transform the formula to a str for the attribute only
            self.formula = pymacroms.getFormulaStrFromCounter(formula)

        # adductIon that is passed must be a str that is present in the database
        self.adductIon = adductIon
        self.charge = charge
        try:
            adductIon = pymacroms.database.ionising_species[adductIon]  # [formula: Counter, charge: int, mass_correction: float]
        except:
            sys.exit("adductIon not found in database!")
        self.amountAdductIons = abs(int(charge / adductIon[1]))
        # add an int amount of ionising species: charge/charge-per-ion
        # again, the internal variable is a counter, the self. variant is a str
        formula_ion = formula + Counter({element: amount*self.amountAdductIons for element, amount in adductIon[0].items()})
        self.formula_ion = pymacroms.getFormulaStrFromCounter(formula_ion)
        # Generate the isotopologues using IsoSpecPy
        isotopologues = IsoSpecPy.IsoThresholdGenerator(formula=self.formula_ion, threshold=minRelAbundance, absolute=False, get_confs=False, use_lprobs=False)
        # IsoSpecPy returns the isotopic distribution as an iterable (mass, 'normalised' abundance (= probability))
        # since IsoSpec generates a list of isotopes with a relative abundance > minimumRelAbundance, further filtering is not necessary
        # total m/z = mass + mass correction (amount of ionising species * correction per ionising specie) devided by charge
        self.isotopicDist = sorted(list(((mass + self.amountAdductIons * adductIon[2])/charge, prob) for mass, prob in isotopologues))
        self.massMin = self.isotopicDist[0][0]
        self.massMax = self.isotopicDist[len(self.isotopicDist)-1][0]
        self.massMostAbundant = sorted(self.isotopicDist, key=itemgetter(1), reverse=True)[0][0]
        self.molecularWeight = pymacroms.getMolecularWeight(pymacroms.getCounterFromFormula(self.formula))

    def printIsotopicDist(self):
        print(self.formula + " --> " + self.formula_ion)
        print("Mass\tRel. Abundance")
        for mass, relAbund in pymacroms.toRelativeAbundance(self.isotopicDist):
            print(str(mass) + "\t" + str(relAbund))

    def plotIsotopicDist(self, color = "black"):
        mz_axis = []
        relAbund_axis = []
        for mz, relAbund in pymacroms.toRelativeAbundance(self.isotopicDist):
            mz_axis += list([mz - 1e-10, mz, mz + 1e-10])
            relAbund_axis += list([-0.1, relAbund, -0.1])
        plt.plot(mz_axis, relAbund_axis, color=color)
        plt.title(r"Isotopic distribution for {} ({} + {}{})".format(re.sub("([0-9]+)", "$_{\\1}$", self.formula_ion), re.sub("([0-9]+)", "$_{\\1}$", self.formula), self.amountAdductIons, re.sub("([0-9]+)", "$_{\\1}$", re.sub("([+-]+)", "$^{\\1}$", self.adductIon)))) # re.sub("([0-9]+)", "$_{\\1}$", self.formula_ion)
        plt.xlabel("m/z")
        plt.ylabel("Rel. abundance")
        plt.ylim(0, 1.05)
        plt.show()

    def inMassRange(self, massRange: list, completelyInRange: bool = False):
        if completelyInRange and self.massMin >= massRange[0] and self.massMax <= massRange[1]:
            return True
        elif not completelyInRange and self.massMax >= massRange[0] and self.massMin <= massRange[1]:
            return True
        else:
            return False

    def hasIsotopeNearMass(self, experimentalMass: float, ppmDev: float = 5):
        for isotope in self.isotopicDist:
            if isclose(isotope[0], experimentalMass, rel_tol=(ppmDev/1e6)):
                return True
        return False

    def getIsotopesNearMass(self, experimentalMass: float, ppmDev: float = 5, resolution: float = None):
        isotopesNearMass = []
        if resolution is None:
            for isotopeMass, normAbundance in self.isotopicDist:
                deviation = abs(experimentalMass - isotopeMass) / isotopeMass * 1e6
                if deviation <= ppmDev:
                    isotopesNearMass.append((isotopeMass, normAbundance, deviation))
        else:
            if not hasattr(self, "resolution"):
                self.resolution = resolution
                self.isotopicDist_resolution = pymacroms.combineIsotopes(self.isotopicDist, resolution)
                self.massMostAbundant = sorted(self.isotopicDist_resolution, key=itemgetter(1), reverse=True)[0][0]
            elif self.resolution != resolution:
                self.resolution = resolution
                self.isotopicDist_resolution = pymacroms.combineIsotopes(self.isotopicDist, resolution)
                self.massMostAbundant = sorted(self.isotopicDist_resolution, key=itemgetter(1), reverse=True)[0][0]
            for isotopeMass, normAbundance in self.isotopicDist_resolution:
                deviation = abs(experimentalMass - isotopeMass) / isotopeMass * 1e6
                if deviation <= ppmDev:
                    isotopesNearMass.append((isotopeMass, normAbundance, deviation))
        if len(isotopesNearMass) > 0:
            return isotopesNearMass
        else:
            return None







# -*- coding: utf-8 -*-
#
# Method asNumpyArray() is copied from RawQuant, copyright (c) 2018 kevinkovalchik. All rights reserved.
#
# Copyright (c) 2019 Kevin De Bruycker and Tim Krappitz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import sys
from collections import OrderedDict
import pandas as pd
import numpy as np
import re
import ctypes
import clr, System
from System import Array, Int32
clr.AddReference('System.Runtime.InteropServices')
if (sys.platform is 'linux') | (sys.platform is 'darwin'):  # check if we are on a Linux or mac OS X system
    clr.AddReference('pymacroms/RawFileReader/Unix/ThermoFisher.CommonCore.Data')
else:
    clr.AddReference('pymacroms/RawFileReader/Windows/ThermoFisher.CommonCore.Data')
    # clr.AddReference('../RawQuant/RawFileReader/ThermoFisher.CommonCore.Data')
from System.Runtime.InteropServices import GCHandle, GCHandleType
from ThermoFisher.CommonCore.Data import Business


class MSData:

    def __init__(self, filename, filterCharge = None, filterScans = None, averageOverScanRange = None, averageOverRetentionTime = None, MSScanFilter = None, mzOffset = 0):

        """

        :type filename: str
        :type filterCharge: int or list
        :type filterScans: list
        :type averageOverScanRange: list
        :type averageOverRetentionTime: list
        :type MSScanFilter: str
        :type mzOffset: float
        """

        self.filename = re.sub("\A(.*)\.([^.]*)\Z", "\\1", filename)
        self.extension = re.sub("\A(.*)\.([^.]*)\Z", "\\2", filename)
        self.mzOffset = float(mzOffset)

        print("Loading spectrum...")
        if self.extension.lower() == "csv":
            self.scans = OrderedDict({'1': {'rt': 0.0, 'spectrum': pd.read_csv(filename, "\t", names=["mz", "abundance"])}})
            self.scans['1']['spectrum'].mz += self.mzOffset
            self.scanRange = range(1, 2)
        elif self.extension.lower() == "raw":
            rawfile = Business.RawFileReaderFactory.ReadFile(filename)
            rawfile.SelectInstrument(0, 1)
            self.scanRange = range(rawfile.RunHeaderEx.FirstSpectrum, rawfile.RunHeaderEx.LastSpectrum + 1)
            self.MSScanFilters = list(MSScanFilter.Filter.ToString() for MSScanFilter in rawfile.GetFilters())
            if len(self.MSScanFilters) > 1 and MSScanFilter is None:
                sys.exit("Multiple MS Scan Filters found. Please provide one of the following to MSData (via "
                         "arg MSScanFilter):\n'" + "'\n'".join(self.MSScanFilters) + "'")
            scanAverager = Business.ScanAveragerFactory.GetScanAverager(rawfile)

            if (averageOverScanRange and averageOverRetentionTime):
                sys.exit("Please provide either averageOverScanRange or averageOverRetentionTime, not both.")
            elif (averageOverScanRange == 'all' or averageOverRetentionTime == 'all'):
                print("--> Averaging all scans...")
                averageScan = scanAverager.GetAverageScanInScanRange(rawfile.RunHeaderEx.FirstSpectrum, rawfile.RunHeaderEx.LastSpectrum, MSScanFilter)
            elif averageOverScanRange:
                print("--> Averaging scans " + str(averageOverScanRange[0]) + " to " + str(averageOverScanRange[1]) + "...")
                averageScan = scanAverager.GetAverageScanInScanRange(averageOverScanRange[0], averageOverScanRange[1], MSScanFilter)
            elif averageOverRetentionTime:
                print("--> Averaging scans with RT from " + str(averageOverRetentionTime[0]) + " to " + str(averageOverRetentionTime[1]) + " minutes...")
                averageScan = scanAverager.GetAverageScanInTimeRange(float(averageOverRetentionTime[0]), float(averageOverRetentionTime[1]), MSScanFilter)
            else:
                averageScan = None

            if averageScan:
                peakData = pd.DataFrame(zip(self.asNumpyArray(averageScan.CentroidScan.Masses),
                                            self.asNumpyArray(averageScan.CentroidScan.Intensities),
                                            self.asNumpyArray(averageScan.CentroidScan.Charges)),
                                        columns=["mz", "abundance", "Charge"])
                peakData.mz += self.mzOffset
                if filterCharge is not None:
                    if type(filterCharge) == int:
                        filterCharge = [filterCharge]
                    peakData.drop(peakData[~peakData["Charge"].isin(filterCharge)].index, inplace=True)
                peakData.drop(columns=["Charge"], inplace=True)
                peakData.reset_index(drop=True, inplace=True)
                self.scans = OrderedDict({'1': {'rt': 0.0, 'spectrum': peakData}})
            else:
                self.scans = OrderedDict()
                for scan in self.scanRange:
                    if MSScanFilter is not None:
                        if rawfile.GetScanStatsForScanNumber(scan).ScanType != MSScanFilter:
                            continue
                    if filterScans is not None:
                        if scan not in filterScans:
                            continue
                    tempData = rawfile.GetCentroidStream(scan, None) # gives access to temp.Masses, .Intensities, .Resolutions, .Baselines, .Noises and .Charges
                    peakData = pd.DataFrame(zip(self.asNumpyArray(tempData.Masses), self.asNumpyArray(tempData.Intensities), self.asNumpyArray(tempData.Charges)), columns=["mz", "abundance", "Charge"])
                    peakData.mz += self.mzOffset
                    if filterCharge is not None:
                        if type(filterCharge) == int:
                            filterCharge = [filterCharge]
                        peakData.drop(peakData[~peakData["Charge"].isin(filterCharge)].index, inplace=True)
                    peakData.drop(columns=["Charge"], inplace=True)
                    peakData.reset_index(drop=True, inplace=True)
                    self.scans[str(scan)] = {'rt': rawfile.RetentionTimeFromScanNumber(scan), 'spectrum': peakData}
        else:
            sys.exit("Filename extension not recognised!")

        print("--> Done\n")

    def asNumpyArray(self, netArray):
        '''
        Given a CLR `System.Array` returns a `numpy.ndarray`.  See _MAP_NET_NP for
        the mapping of CLR types to Numpy dtypes.
        '''
        # _MAP_NP_NET = {
        #     np.dtype('float32'): System.Single,
        #     np.dtype('float64'): System.Double,
        #     np.dtype('int8'): System.SByte,
        #     np.dtype('int16'): System.Int16,
        #     np.dtype('int32'): System.Int32,
        #     np.dtype('int64'): System.Int64,
        #     np.dtype('uint8'): System.Byte,
        #     np.dtype('uint16'): System.UInt16,
        #     np.dtype('uint32'): System.UInt32,
        #     np.dtype('uint64'): System.UInt64,
        #     np.dtype('bool'): System.Boolean,
        # }
        _MAP_NET_NP = {
            'Single': np.dtype('float32'),
            'Double': np.dtype('float64'),
            'SByte': np.dtype('int8'),
            'Int16': np.dtype('int16'),
            'Int32': np.dtype('int32'),
            'Int64': np.dtype('int64'),
            'Byte': np.dtype('uint8'),
            'UInt16': np.dtype('uint16'),
            'UInt32': np.dtype('uint32'),
            'UInt64': np.dtype('uint64'),
            'Boolean': np.dtype('bool'),
        }

        dims = np.empty(netArray.Rank, dtype=int)
        for I in range(netArray.Rank):
            dims[I] = netArray.GetLength(I)
        netType = netArray.GetType().GetElementType().Name

        try:
            npArray = np.empty(dims, order='C', dtype=_MAP_NET_NP[netType])
        except KeyError:
            raise NotImplementedError("asNumpyArray does not yet support System type {}".format(netType) )

        try: # Memmove
            sourceHandle = GCHandle.Alloc(netArray, GCHandleType.Pinned)
            sourcePtr = sourceHandle.AddrOfPinnedObject().ToInt64()
            destPtr = npArray.__array_interface__['data'][0]
            ctypes.memmove(destPtr, sourcePtr, npArray.nbytes)
        finally:
            if sourceHandle.IsAllocated: sourceHandle.Free()
        return npArray



# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Kevin De Bruycker and Tim Krappitz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import absolute_import
import sys
import struct

name = "pymacroms"
version = '1.0.6'

if not hasattr(sys, "version_info") or sys.version_info < (3, 6):
    raise RuntimeError("pymacroms requires Python 3.6 or later.")
if struct.calcsize("P") * 8 != 64:
    raise RuntimeError("pymacroms requires Python 64 bit.")

from . import database
from .single_molecule import Molecule
from .polymer import Polymer
from .spectrum_processing import Spectrum
from .import_spectrum import MSData
from .helper_functions import *

del struct


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pymacroms",
    version="1.0.6",
    author="Kevin De Bruycker and Tim Krappitz",
    author_email="kevindebruycker@gmail.com",
    description="pyMacroMS - High performance quantification of complex high resolution polymer mass spectra",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://macroarc.org/research/macroarc-pyMacroMS.html",
    license="MIT License",
    packages=setuptools.find_packages(),
    package_data={'pymacroms': ['RawFileReader/Unix/*', 'RawFileReader/Windows/*'], },
    install_requires=["IsoSpecPy>=1.9,<2", "matplotlib", "numpy", "pandas", "progressbar2", "pythonnet", "reportlab", "sklearn", "svglib>=0.9.0"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    # python_requires='>=3.6',
)